# -*- coding: utf-8 -*-
# pylint: disable=all
"""
reV offshore wind module command line interface (CLI).

This module uses the NRWAL library to assess offshore losses and LCOE to
complement the simple SAM windpower module.

Everything in this module operates on the native wind resource resolution.
"""
import pprint
import os
import click
import logging
import time

from reV.config.offshore_config import OffshoreConfig
from reV.pipeline.status import Status
from reV.offshore.offshore import Offshore
from reV.utilities.cli_dtypes import SAMFILES, PROJECTPOINTS
from reV import __version__

from rex.utilities.cli_dtypes import STR, INT, STRLIST
from rex.utilities.loggers import init_mult
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import get_class_properties

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--name', '-n', default='reV-off', type=STR,
              help='Job name. Default is "reV-off".')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """reV Offshore Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['VERBOSE'] = verbose


@main.command()
def valid_config_keys():
    """
    Echo the valid Offshore config keys
    """
    click.echo(', '.join(get_class_properties(OffshoreConfig)))


@main.command()
@click.option('--config_file', '-c', required=True,
              type=click.Path(exists=True),
              help='reV offshore configuration json file.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config_file, verbose):
    """Run reV offshore aggregation from a config file."""
    # Instantiate the config object
    config = OffshoreConfig(config_file)
    name = ctx.obj['NAME']

    # take name from config if not default
    if config.name.lower() != 'rev':
        name = config.name
        ctx.obj['NAME'] = name

    # Enforce verbosity if logging level is specified in the config
    if config.log_level == logging.DEBUG:
        verbose = True

    # initialize loggers
    init_mult(name, config.logdir, modules=[__name__, 'reV', 'rex'],
              verbose=verbose)

    # Initial log statements
    logger.info('Running reV offshore aggregation from config '
                'file: "{}"'.format(config_file))
    logger.info('Target output directory: "{}"'.format(config.dirout))
    logger.info('Target logging directory: "{}"'.format(config.logdir))
    logger.debug('The full configuration input is as follows:\n{}'
                 .format(pprint.pformat(config, indent=4)))

    for i, gen_fpath in enumerate(config.parse_gen_fpaths()):
        job_name = '{}_{}'.format(name, str(i).zfill(2))

        if config.execution_control.option == 'local':
            status = Status.retrieve_job_status(config.dirout, 'offshore',
                                                job_name)
            if status != 'successful':
                Status.add_job(
                    config.dirout, 'offshore', job_name, replace=True,
                    job_attrs={'hardware': 'local',
                               'fout': '{}_offshore.h5'.format(job_name),
                               'dirout': config.dirout,
                               'finput': gen_fpath})
                ctx.invoke(direct,
                           gen_fpath=gen_fpath,
                           offshore_fpath=config.offshore_fpath,
                           points=config.project_points,
                           sam_files=config.sam_files,
                           nrwal_configs=config.nrwal_configs,
                           offshore_meta_cols=config.offshore_meta_cols,
                           offshore_nrwal_keys=config.offshore_nrwal_keys,
                           run_all=config.run_all,
                           logdir=config.logdir,
                           verbose=verbose)

        elif config.execution_control.option in ('eagle', 'slurm'):
            ctx.obj['NAME'] = job_name
            ctx.obj['GEN_FPATH'] = gen_fpath
            ctx.obj['OFFSHORE_FPATH'] = config.offshore_fpath
            ctx.obj['PROJECT_POINTS'] = config.project_points
            ctx.obj['SAM_FILES'] = config.sam_files
            ctx.obj['NRWAL_CONFIGS'] = config.nrwal_configs
            ctx.obj['OFFSHORE_META_COLS'] = config.offshore_meta_cols
            ctx.obj['OFFSHORE_NRWAL_KEYS'] = config.offshore_nrwal_keys
            ctx.obj['RUN_ALL'] = config.run_all
            ctx.obj['OUT_DIR'] = config.dirout
            ctx.obj['LOG_DIR'] = config.logdir
            ctx.obj['VERBOSE'] = verbose

            ctx.invoke(slurm,
                       alloc=config.execution_control.allocation,
                       memory=config.execution_control.memory,
                       walltime=config.execution_control.walltime,
                       feature=config.execution_control.feature,
                       module=config.execution_control.module,
                       conda_env=config.execution_control.conda_env)


@main.group(invoke_without_command=True)
@click.option('--gen_fpath', '-gf', type=STR, required=True,
              help='reV wind generation/econ output file.')
@click.option('--offshore_fpath', '-of', type=STR, required=True,
              help='Offshore wind geospatial inputs such as depth and '
              'distance to port. Needs "gid" and "config" columns matching '
              'the project points input.')
@click.option('--points', '-pp', required=True, type=PROJECTPOINTS,
              help='reV project points to analyze. Has to be a string file '
              'path to a project points csv with "gid" and "config" columns. '
              'The config column maps to the sam_files and nrwal_configs '
              'inputs.')
@click.option('--sam_files', '-sf', required=True, type=SAMFILES,
              help='SAM config files lookup mapping config keys to config '
              'filepaths. (required) (dict). Should have the same config '
              'keys as the nrwal_configs input.')
@click.option('--nrwal_configs', '-nc', required=True, type=SAMFILES,
              help='NRWAL config files lookup mapping config keys to config '
              'filepaths. (required) (dict). Should have the same config '
              'keys as the sam_files input.')
@click.option('--offshore_meta_cols', '-mc', default=None, type=STRLIST,
              help='Column labels from offshore_fpath to pass through to the '
              'output meta data. None (default) will use class variable '
              'DEFAULT_META_COLS, and any additional cols requested here will '
              'be added to DEFAULT_META_COLS.')
@click.option('--offshore_nrwal_keys', '-nk', default=None, type=STRLIST,
              help='Keys from the offshore nrwal configs to pass through as '
              'new datasets in the reV output h5.')
@click.option('--log_dir', '-ld', type=STR, default='./logs/',
              help='Directory to save offshore logs.')
@click.option('-ra', '--run_all', is_flag=True,
              help='Flag to run nrwal econ for all sites and ignore the '
              'offshore flag in the meta data.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def direct(ctx, gen_fpath, offshore_fpath, points, sam_files, nrwal_configs,
           offshore_meta_cols, offshore_nrwal_keys, log_dir, run_all, verbose):
    """Main entry point to run offshore wind aggregation"""
    name = ctx.obj['NAME']
    ctx.obj['GEN_FPATH'] = gen_fpath
    ctx.obj['OFFSHORE_FPATH'] = offshore_fpath
    ctx.obj['POINTS'] = points
    ctx.obj['SAM_FILES'] = sam_files
    ctx.obj['NRWAL_CONFIGS'] = nrwal_configs
    ctx.obj['OFFSHORE_META_COLS'] = offshore_meta_cols
    ctx.obj['OFFSHORE_NRWAL_KEYS'] = offshore_nrwal_keys
    ctx.obj['RUN_ALL'] = run_all
    ctx.obj['OUT_DIR'] = os.path.dirname(gen_fpath)
    ctx.obj['LOG_DIR'] = log_dir
    ctx.obj['VERBOSE'] = verbose

    if ctx.invoked_subcommand is None:
        t0 = time.time()
        init_mult(name, log_dir, modules=[__name__, 'reV', 'rex'],
                  verbose=verbose, node=True)

        try:
            Offshore.run(gen_fpath, offshore_fpath, sam_files, nrwal_configs,
                         points, offshore_meta_cols=offshore_meta_cols,
                         offshore_nrwal_keys=offshore_nrwal_keys,
                         run_all=run_all)
        except Exception as e:
            logger.exception('Offshore module failed, received the '
                             'following exception:\n{}'.format(e))
            raise e

        runtime = (time.time() - t0) / 60

        status = {'dirout': os.path.dirname(gen_fpath),
                  'fout': os.path.basename(gen_fpath),
                  'job_status': 'successful',
                  'runtime': runtime, 'finput': gen_fpath}
        Status.make_job_file(os.path.dirname(gen_fpath), 'offshore',
                             name, status)


def get_node_cmd(name, gen_fpath, offshore_fpath, points, sam_files,
                 nrwal_configs, offshore_meta_cols, offshore_nrwal_keys,
                 log_dir, run_all, verbose):
    """Get a CLI call command for the offshore aggregation cli."""

    args = ['-gf {}'.format(SLURM.s(gen_fpath)),
            '-of {}'.format(SLURM.s(offshore_fpath)),
            '-pp {}'.format(SLURM.s(points)),
            '-sf {}'.format(SLURM.s(sam_files)),
            '-nc {}'.format(SLURM.s(nrwal_configs)),
            '-mc {}'.format(SLURM.s(offshore_meta_cols)),
            '-nk {}'.format(SLURM.s(offshore_nrwal_keys)),
            '-ld {}'.format(SLURM.s(log_dir)),
            ]

    if run_all:
        args.append('-ra')
    if verbose:
        args.append('-v')

    cmd = ('python -m reV.offshore.cli_offshore -n {} direct {}'
           .format(SLURM.s(name), ' '.join(args)))
    logger.debug('Creating the following command line call:\n\t{}'.format(cmd))

    return cmd


@direct.command()
@click.option('--alloc', '-a', required=True, type=STR,
              help='SLURM allocation account name.')
@click.option('--feature', '-l', default=None, type=STR,
              help=('Additional flags for SLURM job. Format is "--qos=high" '
                    'or "--depend=[state:job_id]". Default is None.'))
@click.option('--memory', '-mem', default=None, type=INT, help='SLURM node '
              'memory request in GB. Default is None')
@click.option('--walltime', '-wt', default=1.0, type=float,
              help='SLURM walltime request in hours. Default is 1.0')
@click.option('--module', '-mod', default=None, type=STR,
              help='Module to load')
@click.option('--conda_env', '-env', default=None, type=STR,
              help='Conda env to activate')
@click.option('--stdout_path', '-sout', default=None, type=STR,
              help='Subprocess standard output path. Default is in out_dir.')
@click.pass_context
def slurm(ctx, alloc, feature, memory, walltime, module, conda_env,
          stdout_path):
    """slurm (Eagle) submission tool for reV supply curve aggregation."""

    name = ctx.obj['NAME']
    gen_fpath = ctx.obj['GEN_FPATH']
    offshore_fpath = ctx.obj['OFFSHORE_FPATH']
    project_points = ctx.obj['PROJECT_POINTS']
    sam_files = ctx.obj['SAM_FILES']
    nrwal_configs = ctx.obj['NRWAL_CONFIGS']
    offshore_meta_cols = ctx.obj['OFFSHORE_META_COLS']
    offshore_nrwal_keys = ctx.obj['OFFSHORE_NRWAL_KEYS']
    run_all = ctx.obj['RUN_ALL']
    log_dir = ctx.obj['LOG_DIR']
    out_dir = ctx.obj['OUT_DIR']
    verbose = ctx.obj['VERBOSE']

    if stdout_path is None:
        stdout_path = os.path.join(log_dir, 'stdout/')

    cmd = get_node_cmd(name, gen_fpath, offshore_fpath, project_points,
                       sam_files, nrwal_configs, offshore_meta_cols,
                       offshore_nrwal_keys, log_dir, run_all, verbose)
    slurm_manager = ctx.obj.get('SLURM_MANAGER', None)
    if slurm_manager is None:
        slurm_manager = SLURM()
        ctx.obj['SLURM_MANAGER'] = slurm_manager

    status = Status.retrieve_job_status(out_dir, 'offshore', name,
                                        hardware='eagle',
                                        subprocess_manager=slurm_manager)

    if status == 'successful':
        msg = ('Job "{}" is successful in status json found in "{}", '
               'not re-running.'
               .format(name, out_dir))
    elif 'fail' not in str(status).lower() and status is not None:
        msg = ('Job "{}" was found with status "{}", not resubmitting'
               .format(name, status))
    else:
        logger.info('Running reV offshore aggregation on SLURM with '
                    'node name "{}"'.format(name))
        out = slurm_manager.sbatch(cmd,
                                   alloc=alloc,
                                   memory=memory,
                                   walltime=walltime,
                                   feature=feature,
                                   name=name,
                                   stdout_path=stdout_path,
                                   conda_env=conda_env,
                                   module=module)[0]
        if out:
            msg = ('Kicked off reV offshore job "{}" '
                   '(SLURM jobid #{}).'
                   .format(name, out))
            Status.add_job(
                out_dir, 'offshore', name, replace=True,
                job_attrs={'job_id': out, 'hardware': 'eagle',
                           'fout': '{}.csv'.format(name), 'dirout': out_dir})

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running reV offshore CLI.')
        raise



# -*- coding: utf-8 -*-
"""reV tech mapping framework.

This module manages the exclusions-to-resource mapping.
The core of this module is a parallel cKDTree.

Created on Fri Jun 21 16:05:47 2019

@author: gbuster
"""
from concurrent.futures import as_completed
import h5py
import logging
from math import ceil
import numpy as np
import os
from scipy.spatial import cKDTree
from warnings import warn

from reV.supply_curve.extent import SupplyCurveExtent
from reV.utilities.exceptions import FileInputWarning, FileInputError

from rex.resource import Resource
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.utilities import res_dist_threshold

logger = logging.getLogger(__name__)


class TechMapping:
    """Framework to create map between tech layer (exclusions), res, and gen"""

    def __init__(self, excl_fpath, res_fpath, sc_resolution=2560,
                 dist_margin=1.05):
        """
        Parameters
        ----------
        excl_fpath : str
            Filepath to exclusions h5 file, must contain latitude and longitude
            arrays to allow for mapping to resource points
        res_fpath : str
            Filepath to .h5 resource file that we're mapping to.
        sc_resolution : int | None, optional
            Supply curve resolution, does not affect the exclusion to resource
            (tech) mapping, but defines how many exclusion pixels are mapped
            at a time, by default 2560
        dist_margin : float, optional
            Extra margin to multiply times the computed distance between
            neighboring resource points, by default 1.05
        """
        self._excl_fpath = excl_fpath
        self._check_fout()

        self._tree, self._dist_thresh = \
            self._build_tree(res_fpath, dist_margin=dist_margin)

        with SupplyCurveExtent(self._excl_fpath,
                               resolution=sc_resolution) as sc:
            self._sc_resolution = sc.resolution
            self._gids = np.array(list(range(len(sc))), dtype=np.uint32)
            self._excl_shape = sc.exclusions.shape
            self._n_excl = np.product(self._excl_shape)
            self._sc_row_indices = sc.row_indices
            self._sc_col_indices = sc.col_indices
            self._excl_row_slices = sc.excl_row_slices
            self._excl_col_slices = sc.excl_col_slices
            logger.info('Initialized TechMapping object with {} calc chunks '
                        'for {} tech exclusion points'
                        .format(len(self._gids), self._n_excl))

    @property
    def distance_threshold(self):
        """Get the upper bound on NN distance between excl and res points.

        Returns
        -------
        float
            Estimate the distance between resource points. Calculated as half
            of the diagonal between closest resource points, with desired
            extra margin
        """
        return self._dist_thresh

    @staticmethod
    def _build_tree(res_fpath, dist_margin=1.05):
        """
        Build cKDTree from resource lat, lon coordinates. Compute minimum
        intra point distance between resource gids with provided extra margin.

        Parameters
        ----------
        res_fpath : str
            Filepath to .h5 resource file that we're mapping to.
        dist_margin : float, optional
            Extra margin to multiply times the computed distance between
            neighboring resource points, by default 1.05

        Returns
        -------
        tree : cKDTree
            cKDTree built from resource lat, lon coordinates
        dist_tresh : float
            Estimate the distance between resource points. Calculated as half
            of the diagonal between closest resource points, with desired
            extra margin
        """
        with Resource(res_fpath) as f:
            lat_lons = f.lat_lon

        # pylint: disable=not-callable
        tree = cKDTree(lat_lons)

        dist_thresh = res_dist_threshold(lat_lons, tree=tree,
                                         margin=dist_margin)

        return tree, dist_thresh

    @staticmethod
    def _make_excl_iarr(shape):
        """
        Create 2D array of 1D index values for the flattened h5 excl extent

        Parameters
        ----------
        shape : tuple
            exclusion extent shape

        Returns
        -------
        iarr : ndarray
            2D array of 1D index values for the flattened h5 excl extent
        """
        iarr = np.arange(np.product(shape), dtype=np.uint32)

        return iarr.reshape(shape)

    @staticmethod
    def _get_excl_slices(gid, sc_row_indices, sc_col_indices, excl_row_slices,
                         excl_col_slices):
        """
        Get the row and column slices of the exclusions grid corresponding
        to the supply curve point gid.

        Parameters
        ----------
        gid : int
            Supply curve point gid.
        sc_row_indices : list
            List of row indices in exclusion array for for every sc_point gid
        sc_col_indices : list
            List of column indices in exclusion array for for every sc_point
            gid
        excl_row_slices : list
            List representing the supply curve points rows. Each list entry
            contains the exclusion row slice that are included in the sc
            point.
        excl_col_slices : list
            List representing the supply curve points columns. Each list entry
            contains the exclusion columns slice that are included in the sc
            point.

        Returns
        -------
        row_slice : int
            Exclusions grid row index slice corresponding to the sc point gid.
        col_slice : int
            Exclusions grid col index slice corresponding to the sc point gid.
        """

        row_slice = excl_row_slices[sc_row_indices[gid]]
        col_slice = excl_col_slices[sc_col_indices[gid]]

        return row_slice, col_slice

    @classmethod
    def _get_excl_coords(cls, excl_fpath, gids, sc_row_indices, sc_col_indices,
                         excl_row_slices, excl_col_slices,
                         coord_labels=('latitude', 'longitude')):
        """
        Extract the exclusion coordinates for teh desired gids for TechMapping.

        Parameters
        ----------
        gids : np.ndarray
            Supply curve gids with tech exclusion points to map to the
            resource meta points.
        excl_fpath : str
            Filepath to exclusions h5 file, must contain latitude and longitude
            arrays to allow for mapping to resource points
        sc_row_indices : list
            List of row indices in exclusion array for for every sc_point gid
        sc_col_indices : list
            List of column indices in exclusion array for for every sc_point
            gid
        excl_row_slices : list
            List representing the supply curve points rows. Each list entry
            contains the exclusion row slice that are included in the sc
            point.
        excl_col_slices : list
            List representing the supply curve points columns. Each list entry
            contains the exclusion columns slice that are included in the sc
            point.
        coord_labels : tuple
            Labels for the coordinate datasets.

        Returns
        -------
        coords_out : list
            List of arrays of the un-projected latitude, longitude array of
            tech exclusion points. List entries correspond to input gids.
        """
        coords_out = []
        with h5py.File(excl_fpath, 'r') as f:
            for gid in gids:
                row_slice, col_slice = cls._get_excl_slices(gid,
                                                            sc_row_indices,
                                                            sc_col_indices,
                                                            excl_row_slices,
                                                            excl_col_slices)
                try:
                    lats = f[coord_labels[0]][row_slice, col_slice]
                    lons = f[coord_labels[1]][row_slice, col_slice]
                    emeta = np.vstack((lats.flatten(), lons.flatten())).T
                except Exception as e:
                    m = ('Could not unpack coordinates for gid {} with '
                         'row/col slice {}/{}. Received the following '
                         'error:\n{}'.format(gid, row_slice, col_slice, e))
                    logger.error(m)
                    raise e

                coords_out.append(emeta)

        return coords_out

    @classmethod
    def map_resource_gids(cls, gids, excl_fpath, sc_row_indices,
                          sc_col_indices, excl_row_slices, excl_col_slices,
                          tree, dist_thresh):
        """Map exclusion gids to the resource meta.

        Parameters
        ----------
        gids : np.ndarray
            Supply curve gids with tech exclusion points to map to the
            resource meta points.
        excl_fpath : str
            Filepath to exclusions h5 file, must contain latitude and longitude
            arrays to allow for mapping to resource points
        sc_row_indices : list
            List of row indices in exclusion array for for every sc_point gid
        sc_col_indices : list
            List of column indices in exclusion array for for every sc_point
            gid
        excl_row_slices : list
            List representing the supply curve points rows. Each list entry
            contains the exclusion row slice that are included in the sc
            point.
        excl_col_slices : list
            List representing the supply curve points columns. Each list entry
            contains the exclusion columns slice that are included in the sc
            point.
        tree : cKDTree
            cKDTree built from resource lat, lon coordinates
        dist_tresh : float
            Estimate the distance between resource points. Calculated as half
            of the diagonal between closest resource points, with an extra
            5% margin

        Returns
        -------
        ind : list
            List of arrays of index values from the NN. List entries correspond
            to input gids.
        """
        logger.debug('Getting tech map coordinates for chunks {} through {}'
                     .format(gids[0], gids[-1]))
        ind_out = []
        coords_out = cls._get_excl_coords(excl_fpath, gids, sc_row_indices,
                                          sc_col_indices, excl_row_slices,
                                          excl_col_slices)

        logger.debug('Running tech mapping for chunks {} through {}'
                     .format(gids[0], gids[-1]))
        for i, _ in enumerate(gids):
            dist, ind = tree.query(coords_out[i])
            ind[(dist >= dist_thresh)] = -1
            ind_out.append(ind)

        return ind_out

    @staticmethod
    def save_tech_map(excl_fpath, dset, indices, distance_threshold=None,
                      res_fpath=None, chunks=(128, 128)):
        """Save tech mapping indices and coordinates to an h5 output file.

        Parameters
        ----------
        excl_fpath : str
            Filepath to exclusions h5 file to add techmap to as 'dset'
        dset : str
            Dataset name in fpath_out to save mapping results to.
        indices : np.ndarray
            Index values of the NN resource point. -1 if no res point found.
            2D integer array with shape equal to the exclusions extent shape.
        distance_threshold : float
            Distance upper bound to save as attr.
        res_fpath : str, optional
            Filepath to .h5 resource file that we're mapping to,
            by default None
        chunks : tuple
            Chunk shape of the 2D output datasets.
        """
        logger.info('Writing tech map "{}" to {}'.format(dset, excl_fpath))

        shape = indices.shape
        chunks = (np.min((shape[0], chunks[0])), np.min((shape[1], chunks[1])))

        with h5py.File(excl_fpath, 'a') as f:
            if dset in list(f):
                wmsg = ('TechMap results dataset "{}" is being replaced '
                        'in pre-existing Exclusions TechMapping file "{}"'
                        .format(dset, excl_fpath))
                logger.warning(wmsg)
                warn(wmsg, FileInputWarning)
                f[dset][...] = indices
            else:
                f.create_dataset(dset, shape=shape, dtype=indices.dtype,
                                 data=indices, chunks=chunks)

            if distance_threshold:
                f[dset].attrs['distance_threshold'] = distance_threshold

            if res_fpath:
                f[dset].attrs['src_res_fpath'] = res_fpath

        logger.info('Successfully saved tech map "{}" to {}'
                    .format(dset, excl_fpath))

    def _check_fout(self):
        """Check the TechMapping output file for cached data."""
        with h5py.File(self._excl_fpath, 'r') as f:
            if 'latitude' not in f or 'longitude' not in f:
                emsg = ('Datasets "latitude" and/or "longitude" not in '
                        'pre-existing Exclusions TechMapping file "{}". '
                        'Cannot proceed.'
                        .format(os.path.basename(self._excl_fpath)))
                logger.exception(emsg)
                raise FileInputError(emsg)

    def map_resource(self, max_workers=None, points_per_worker=10):
        """
        Map all resource gids to exclusion gids

        Parameters
        ----------
        max_workers : int, optional
            Number of cores to run mapping on. None uses all available cpus,
            by default None
        points_per_worker : int, optional
            Number of supply curve points to map to resource gids on each
            worker, by default 10

        Returns
        -------
        indices : np.ndarray
            Index values of the NN resource point. -1 if no res point found.
            2D integer array with shape equal to the exclusions extent shape.
        """
        gid_chunks = ceil(len(self._gids) / points_per_worker)
        gid_chunks = np.array_split(self._gids, gid_chunks)

        # init full output arrays
        indices = -1 * np.ones((self._n_excl, ), dtype=np.int32)
        iarr = self._make_excl_iarr(self._excl_shape)

        futures = {}
        loggers = [__name__, 'reV']
        with SpawnProcessPool(max_workers=max_workers,
                              loggers=loggers) as exe:

            # iterate through split executions, submitting each to worker
            for i, gid_set in enumerate(gid_chunks):
                # submit executions and append to futures list
                futures[exe.submit(self.map_resource_gids,
                                   gid_set,
                                   self._excl_fpath,
                                   self._sc_row_indices,
                                   self._sc_col_indices,
                                   self._excl_row_slices,
                                   self._excl_col_slices,
                                   self._tree,
                                   self.distance_threshold)] = i

            n_finished = 0
            for future in as_completed(futures):
                n_finished += 1
                logger.info('Parallel TechMapping futures collected: '
                            '{} out of {}'
                            .format(n_finished, len(futures)))

                i = futures[future]
                result = future.result()

                for j, gid in enumerate(gid_chunks[i]):
                    row_slice, col_slice = self._get_excl_slices(
                        gid,
                        self._sc_row_indices,
                        self._sc_col_indices,
                        self._excl_row_slices,
                        self._excl_col_slices)
                    ind_slice = iarr[row_slice, col_slice].flatten()
                    indices[ind_slice] = result[j]

        indices = indices.reshape(self._excl_shape)

        return indices

    @classmethod
    def run(cls, excl_fpath, res_fpath, dset=None, sc_resolution=2560,
            dist_margin=1.05, max_workers=None, points_per_worker=10):
        """Run parallel mapping and save to h5 file.

        Parameters
        ----------
        excl_fpath : str
            Filepath to exclusions h5 (tech layer). dset will be
            created in excl_fpath.
        res_fpath : str
            Filepath to .h5 resource file that we're mapping to.
        dset : str, optional
            Dataset name in excl_fpath to save mapping results to, if None
            do not save tech_map to excl_fpath, by default None
        sc_resolution : int | None, optional
            Supply curve resolution, does not affect the exclusion to resource
            (tech) mapping, but defines how many exclusion pixels are mapped
            at a time, by default 2560
        dist_margin : float, optional
            Extra margin to multiply times the computed distance between
            neighboring resource points, by default 1.05
        max_workers : int, optional
            Number of cores to run mapping on. None uses all available cpus,
            by default None
        points_per_worker : int, optional
            Number of supply curve points to map to resource gids on each
            worker, by default 10

        Returns
        -------
        indices : np.ndarray
            Index values of the NN resource point. -1 if no res point found.
            2D integer array with shape equal to the exclusions extent shape.
        """
        kwargs = {"dist_margin": dist_margin,
                  "sc_resolution": sc_resolution}
        mapper = cls(excl_fpath, res_fpath, **kwargs)
        indices = mapper.map_resource(max_workers=max_workers,
                                      points_per_worker=points_per_worker)

        if dset:
            mapper.save_tech_map(excl_fpath, dset, indices,
                                 distance_threshold=mapper.distance_threshold,
                                 res_fpath=res_fpath)

        return indices



# -*- coding: utf-8 -*-
"""Curtailment utility methods.

Created on Fri Mar  1 13:47:30 2019

@author: gbuster
"""
import datetime
import logging
import numpy as np
import pandas as pd
from warnings import warn

from reV.utilities.exceptions import HandlerWarning

from rex.utilities.solar_position import SolarPosition
from rex.utilities.utilities import check_tz, get_lat_lon_cols

logger = logging.getLogger(__name__)


def curtail(resource, curtailment, random_seed=0):
    """Curtail the SAM wind resource object based on project points.

    Parameters
    ----------
    resource : rex.sam_resource.SAMResource
        SAM resource object for WIND resource.
    curtailment : reV.config.curtailment.Curtailment
        Curtailment config object.
    random_seed : int | NoneType
        Number to seed the numpy random number generator. Used to generate
        reproducable psuedo-random results if the probability of curtailment
        is not set to 1. Numpy random will be seeded with the system time if
        this is None.

    Returns
    -------
    resource : reV.handlers.sam_resource.SAMResource
        Same as the input argument but with the wind speed dataset set to zero
        where curtailment is in effect.
    """

    shape = resource.shape

    # start with curtailment everywhere
    curtail_mult = np.zeros(shape)

    if curtailment.date_range is not None:
        year = resource.time_index.year[0]
        d0 = pd.to_datetime(datetime.datetime(
            month=int(curtailment.date_range[0][:2]),
            day=int(curtailment.date_range[0][2:]),
            year=year), utc=True)
        d1 = pd.to_datetime(datetime.datetime(
            month=int(curtailment.date_range[1][:2]),
            day=int(curtailment.date_range[1][2:]),
            year=year), utc=True)
        time_index = check_tz(resource.time_index)
        mask = (time_index >= d0) & (time_index < d1)
        mask = np.tile(np.expand_dims(mask, axis=1), shape[1])
        curtail_mult = np.where(mask, curtail_mult, 1)

    elif curtailment.months is not None:
        # Curtail resource when in curtailment months
        mask = np.isin(resource.time_index.month, curtailment.months)
        mask = np.tile(np.expand_dims(mask, axis=1), shape[1])
        curtail_mult = np.where(mask, curtail_mult, 1)

    else:
        msg = ('You must specify either months or date_range over '
               'which curtailment is possible!')
        logger.error(msg)
        raise KeyError(msg)

    # Curtail resource when curtailment is possible and is nighttime
    lat_lon_cols = get_lat_lon_cols(resource.meta)
    solar_zenith_angle = SolarPosition(
        resource.time_index,
        resource.meta[lat_lon_cols].values).zenith
    mask = (solar_zenith_angle > curtailment.dawn_dusk)
    curtail_mult = np.where(mask, curtail_mult, 1)

    # Curtail resource when curtailment is possible and not raining
    if curtailment.precipitation is not None:
        if 'precipitationrate' not in resource._res_arrays:
            warn('Curtailment has a precipitation threshold of "{}", but '
                 '"precipitationrate" was not found in the SAM resource '
                 'variables. The following resource variables were '
                 'available: {}.'
                 .format(curtailment.precipitation,
                         list(resource._res_arrays.keys())),
                 HandlerWarning)
        else:
            mask = (resource._res_arrays['precipitationrate']
                    < curtailment.precipitation)
            curtail_mult = np.where(mask, curtail_mult, 1)

    # Curtail resource when curtailment is possible and temperature is high
    if curtailment.temperature is not None:
        mask = (resource._res_arrays['temperature']
                > curtailment.temperature)
        curtail_mult = np.where(mask, curtail_mult, 1)

    # Curtail resource when curtailment is possible and not that windy
    if curtailment.wind_speed is not None:
        mask = (resource._res_arrays['windspeed']
                < curtailment.wind_speed)
        curtail_mult = np.where(mask, curtail_mult, 1)

    if curtailment.equation is not None:
        # pylint: disable=W0123,W0612
        wind_speed = resource._res_arrays['windspeed']
        temperature = resource._res_arrays['temperature']
        if 'precipitationrate' in resource._res_arrays:
            precipitation_rate = resource._res_arrays['precipitationrate']
        mask = eval(curtailment.equation)
        curtail_mult = np.where(mask, curtail_mult, 1)

    # Apply probability mask when curtailment is possible.
    if curtailment.probability != 1:
        np.random.seed(seed=random_seed)
        mask = np.random.rand(shape[0], shape[1]) < curtailment.probability
        curtail_mult = np.where(mask, curtail_mult, 1)

    # Apply curtailment multiplier directly to resource
    resource.curtail_windspeed(resource.sites, curtail_mult)

    return resource



# -*- coding: utf-8 -*-
"""Slotted memory framework classes."""


class SlottedDict:
    """Slotted memory dictionary emulator."""

    # make attribute slots for all dictionary keys
    __slots__ = ['var_list']

    def __init__(self):
        self.var_list = []

    def __setitem__(self, key, value):
        """Send data to a slot. Raise KeyError if key is not recognized"""
        if key in self.__slots__:
            if key not in self.var_list:
                self.var_list.append(key)
            setattr(self, key, value)
        else:
            raise KeyError('Could not save "{}" to slotted dictionary. '
                           'The following output variable slots are '
                           'available: {}'.format(key, self.__slots__))

    def __getitem__(self, key):
        """Retrieve data from slot. Raise KeyError if key is not recognized"""
        if key in self.var_list:
            return getattr(self, key)
        else:
            raise KeyError('Variable "{}" has not been saved to this slotted '
                           'dictionary instance. Saved variables are: {}'
                           .format(key, self.keys()))

    def update(self, slotted_dict):
        """Add output variables from another instance into this instance.

        Parameters
        ----------
        slotted_dict : SlottedDict
            An different instance of this class (slotted dictionary class) to
            merge into this instance. Variable data in this instance could be
            overwritten by the new data.
        """

        attrs = slotted_dict.var_list
        for attr in attrs:
            if attr in self.__slots__:
                value = getattr(slotted_dict, attr, None)
                if value is not None:
                    self[attr] = value

    def items(self):
        """Get an items iterator similar to a dictionary.

        Parameters
        ----------
        items : iterator
            [key, value] iterator similar to the output of dict.items()
        """

        keys = self.keys()
        values = self.values()
        return zip(keys, values)

    def keys(self):
        """Get a keys list similar to a dictionary.

        Parameters
        ----------
        key : list
            List of slotted variable names that have been set.
        """
        return self.var_list

    def values(self):
        """Get a values list similar to a dictionary.

        Parameters
        ----------
        values : list
            List of slotted variable values that have been set.
        """
        return [self[k] for k in self.var_list]
